#ifndef Counter_h
#define Counter_h

#include "core/Solver.h"

#include "counter/Config.h"
#include "counter/Instance.h"
#include "preprocessor/Preprocessor.h"
#include "counter/Cache.h"

using namespace Glucose;

namespace GraphPC {

/* ****************
 * Decision
 ******************/
class Decision {
	int trail_pos_;
	bool cur_branch_;

	unsigned basepos_;

	mpz_class models_[2];
	bool hasmodel_[2];

public:
	Decision(int trail_pos, unsigned basepos) :
		trail_pos_(trail_pos), cur_branch_(false), basepos_(basepos), models_{ 0, 0 }, hasmodel_{false, false} {
	}

	int trailPos() {
		return trail_pos_;
	}
	unsigned basePos() {
		return basepos_;
	}

	bool isFirstBranch() {
		return !cur_branch_;
	}
	void changeBranch() {
		cur_branch_ = true;
	}

	// Methods on models
	const mpz_class totalModels() const {
		return models_[0] + models_[1];
	}

	const bool hasModel() const {
		return hasmodel_[false] || hasmodel_[true];
	}

	const mpz_class currentBranchModels() const {
		return models_[cur_branch_];
	}

	bool isUnSAT() {
		return !hasmodel_[cur_branch_];
	}

	void increaseModels(const mpz_class &nmodels, bool found = true) {
		models_[cur_branch_] = found ? nmodels : 0;
		hasmodel_[cur_branch_] = found;
	}
};

/* ****************
 * Counter
 * ****************/
#define GraphPC_VERSION "GraphPathCount-0.1"

enum progressT {
	INIT, LOADED, PREPROCESSED, COMPLETED_BYPP, COMPLETED, FAILED
};

enum btStateT {
	EXIT, RESOLVED
};

class Counter: public Solver {
public:
	// Constructor/Destructor
	//
	Counter(Configuration& config_);
	~Counter() { };

	// Methods
	//
	void load(std::istream& in, std::istream &map);	// load input files
	bool preprocess();									// preprocessing for simplification
	bool countPaths();		    						// main count method

	const mpz_class& getCount() { return npaths; }	// get the computed count
	const bool isSat() 			{ return sat; }		// sat or unsat

	void printStats() const;								// print statistics

protected:
	// Data members
	PPMC::Preprocessor pp;	// preprocessor
	Instance ins;				// CNF instance (temporal)

	// cache
	Cache cache;
	BitVec key;

	std::vector<Decision> dl_;	// extra decision stack for counting
	vec<int> nodeTrail;			// keep the current path

	vec<Lit> unitcls;			// The list of learnt unit clauses

	int limlevel;				// for limited backjump
	int fixed_level_;

	bool on_bj;				// limited bachjump on/off
	bool on_simp;				// in-processing simplification on/off
	bool on_restart;			// restarting on/off

	progressT progress;		// counter progress

	bool setStart;
	Glucose::Var startvar;

	// result
	bool sat;
	mpz_class npaths;

	// options
	ConfigCounter config;
	int verbosity_c;			// verbosity level

	/// statistics
	int npvars_isolated;
	uint64_t nbackjumps, nbackjumps_sp, nrestarts;
	uint64_t reduce_dbs_pre, simp_dbs;
	double simplify_time;

	// *****************
	// Methods
	// *****************

	// import a CNF instance
	void import();

	// initialize
	void init();

	/// count main method
	void count_main(int startnode, int endnode);
	void count_main_all();

	/// variable selection (finding a next edge)
	Var findNextEdge();

	/// backtrack
	btStateT backtrack();
	void cancelCurDL();
	void bjResolve(int level);
	btStateT Resolve(int bk_level, CRef cr, Lit lit);

	bool analyzeMC(CRef confl, vec<Lit>& out_learnt, vec<Lit> & selectors,
			int& out_btlevel, unsigned int &nblevels,
			unsigned int &szWithoutSelectors);
	unsigned int computeLBDMC(const vec<Lit> & lits, int end);
	unsigned int computeLBDMC(const Clause &c);

	// Decision Stack (extra)
	Decision& topDecision() { return dl_.back(); }
	Decision& prevDecision() { return *(dl_.end() - 2); }
	void pushDecision(int trailp) {
		dl_.push_back(Decision(trailp, nodeTrail.size()-1));
	}
	void popDecision() {
		nodeTrail.shrink(nodeTrail.size()-topDecision().basePos());
		dl_.pop_back();
	}

	bool checkfixedDL() {
		if(fixed_level_+1 == dl_.size()) {
			if(!dl_.back().isFirstBranch()) {
				fixed_level_++;
				return true;
			}
		}
		return false;
	}

	bool simplify();

	void initCache();
	void genkey(BitVec &key);
};
}

#endif
