#include "counter/Instance.h"

#include "mtl/Sort.h"

#include <iostream>
#include <fstream>
#include <vector>
#include <algorithm>

using namespace Glucose;
using namespace GraphPC;
using namespace std;

Instance::Instance() :
vars (0),
npvars (0),
freevars (0),
nNodes_ (0),
nEdges_ (0),
start (0),
end (0),
maxpathlen (0),
allpairs(false),
unsat (false)
{}

static inline Lit SignedIntToLit(int signed_int) {
	assert(signed_int != 0);
	int v = abs(signed_int)-1;
	return (signed_int > 0) ? mkLit(v) : ~mkLit(v);
}

static void Tokens(string buf, vector<string>& ret) {
	ret.push_back("");
	for (char c : buf) {
		if (std::isspace(c)) {
			if (!ret.back().empty()) {
				ret.push_back("");
			}
		} else {
			ret.back() += c;
		}
	}
	if (ret.back().empty()) ret.pop_back();
}

void Instance::loadCNF(istream& in) {
	string buf;
	vector<Lit> ps;
	vector<string> tokens;

	assignedLits.clear();
	learnts.clear();

	while (getline(in, buf)) {
		if (buf.empty()) continue;
		tokens.clear();
		Tokens(buf, tokens);

		if(tokens.size() == 0) continue;
		else if (tokens[0] == "p") { // Read header
			if (tokens.size() == 4 && tokens[1] == "cnf") {
				vars = stoi(tokens[2]);
				assigns.resize(vars, l_Undef);
			}
			else
				cerr << "c c Header Error!" << endl;
		}
		else if (tokens[0][0] == 'c') {
			// Read comment
		}
		else { // Read clause
			int lit = 0;
			assert(tokens.size() > 0);
			for(auto token : tokens) {
				lit = stoi(token);
				if(lit == 0) break;
				ps.push_back(SignedIntToLit(lit));
			}
			if(lit == 0) {
				if(!addClause(ps)) {
					return;
				}
				ps.clear();
			} else {
				cerr << "c c invalid Clause!" << endl;
			}
		}
	}
}

void Instance::loadGraph(istream& in, bool allpairs) {
	string buf;
	vector<Lit> ps;
	vector<string> tokens;

	this->allpairs = allpairs;

	while (getline(in, buf)) {
		if (buf.empty()) continue;
		tokens.clear();
		Tokens(buf, tokens);

		if(tokens.size() == 0) continue;
		else if (tokens[0] == "p") { // Read header
			if (tokens.size() == 4 && tokens[1] == "map") {
				nNodes_ = stoi(tokens[2]);

				ispvars.clear();
				ispvars.resize(vars, false);

				if(allpairs) {
					startnodes.clear();
					startnodes.resize(nNodes_);
				}

				edges.clear();
			}
			else
				cerr << "c c Header Error!" << endl;
		}
		else if (tokens[0][0] == 't') {
			// Read Terminals
			start = stoi(tokens[1])-1;
			end = stoi(tokens[2])-1;
		}
		else if (tokens[0][0] == 'l') {
			// Read length
			maxpathlen = stoi(tokens[1]);
		}
		else if (tokens[0][0] == 'c') {
			// Read comment
		}
		else if (tokens[0][0] == 's') {
			int cnfvar = stoi(tokens[1])-1;
			int node = stoi(tokens[2])-1;
			ispvars[cnfvar] = true;
			startnodes[node] = cnfvar;
		}
		else { // Read a correspondence (edgeVar head tail)
			assert(tokens.size() == 3);

			int edgeVar = stoi(tokens[0])-1;
			int tail = stoi(tokens[1])-1;
			int head = stoi(tokens[2])-1;

			// assert(edgeVar < nEdges());
			assert(head < nNodes());
			assert(tail < nNodes());
			assert(head != tail);

			// pvars.push_back(edgeVar);
			ispvars[edgeVar] = true;
			edges.push_back(mkEdge(edgeVar, tail, head));
		}
	}

	nEdges_ = edges.size();
	npvars = allpairs ? nNodes_+nEdges_ : nEdges_;
	assert(npvars <= vars);
}

bool Instance::addClause(vector<Lit>& ps, bool learnt) {
	sort(ps.begin(), ps.end());

	vec<Lit> oc;
	oc.clear();

	Lit p; int i, j, flag = 0;

	for (i = j = 0, p = lit_Undef; i < ps.size(); i++)
		if (value(ps[i]) == l_True || ps[i] == ~p)
			return true;
		else if (value(ps[i]) != l_False && ps[i] != p)
			ps[j++] = p = ps[i];
	ps.resize(j);

	if (ps.size() == 0) {
		unsat = true;
		return false;
	}
	else if (ps.size() == 1) {
		assigns[var(ps[0])] = lbool(!sign(p));
		assignedLits.push_back(ps[0]);
		// No check by UP here. Preprocessor will do later.
	}
	else {
		clauses.push_back({});
		copy(ps.begin(), ps.end(), back_inserter(clauses.back()));
	}
	return true;
}

// For Debug
 void Instance::toDimacs(std::ostream& out)
 {
	using namespace std;

	out << "c free vars: " << freevars << endl;
	out << "p cnf " << vars + freevars << " " << clauses.size() << endl;

	out << "c p show ";
	for (int i=0; i<npvars; i++)
		out << (i+1) << " ";
	out << "0" << endl;

	for(auto c : clauses) {
		for (auto l : c)
			out << (sign(l) ? "-":"") << (var(l)+1) << " ";
		out << "0" << endl;
	}

	// For Debug
	/*
	for(auto c : learnts) {
		for (auto l : c)
			out << (sign(l) ? "-":"") << (var(l)+1) << " ";
		out << "0" << endl;
	}
	*/
}

 void Instance::printGraphStats(const char* msg) const {
	 cout << "c o Graph Stat[" << string(msg) << "]: " << nNodes_ << " nodes, " << nEdges_ << " edges, maxpathlen=" << maxpathlen << endl;
 }

 void Instance::writeGraph(std::ostream& out)
 {
	 using namespace std;

	 out << "p di_edge " << this->nNodes_ << " " << this->nEdges() << endl;

	 int i = 1;
	 for (auto e : edges) {
		 out << i << " " << e.from << " " << e.to << endl;
		 i++;
	 }

	 out << "t " << start << " " << end << endl;
	 out << "l " << this->maxpathlen << endl;
 }
