#ifndef Instance_h
#define Instance_h

#include <istream>
#include <vector>
#include "core/SolverTypes.h"

namespace GraphPC {

// Use this as a constructor:
struct Edge {
	Glucose::Var cnfvar;
	int from;
	int to;

   // Use this as a constructor:
   friend Edge mkEdge(int from, int to);

   bool operator == (Edge e) const { return cnfvar == e.cnfvar && from == e.from && to == e.to; }
   bool operator != (Edge e) const { return cnfvar != e.cnfvar || from != e.from || to != e.to; }
   bool operator <  (Edge e) const { return cnfvar < e.cnfvar; } // return from < e.from || (from == e.from && to < e.to);  }
};

inline Edge mkEdge (Glucose::Var cnfvar, int v1, int v2) { Edge e; e.cnfvar = cnfvar; e.from = v1; e.to = v2; return e; }

class Instance {
public:
	Instance();

	void loadCNF		(std::istream& in);
	bool addClause	(std::vector<Glucose::Lit>& lits, bool learnt=false);

	void loadGraph	(std::istream& in, bool allpairs);

	size_t nNodes() const;
	size_t nEdges() const;

	Glucose::lbool value (Glucose::Var x) const;
	Glucose::lbool value (Glucose::Lit p) const;

	// graph statistics
	void printGraphStats(const char* msg) const;

	// For Debug
	void toDimacs(std::ostream& out);
	void writeGraph(std::ostream& out);

	// CNF formula
	int vars;
	std::vector<std::vector<Glucose::Lit>> clauses;
	std::vector<std::vector<Glucose::Lit>> learnts;

	// For edges vars (projected)
	int npvars;
	std::vector<bool> ispvars;
	std::vector<Glucose::Var> pvars;

	// Instance keeps temporal fixed literals. Preprocessor will eliminate fixed variables.
	std::vector<Glucose::lbool>	assigns;
	std::vector<Glucose::Lit>	assignedLits;

	// additional information
	int freevars;

	// information about correspondence with a given CNF
	std::vector<Glucose::Lit> fixedLits;						// fixed Literals

	// For Graph
	int nNodes_;
	int nEdges_;
	std::vector<Edge> edges;
	std::vector<Glucose::Var> HeadNodeOf;			// edge -> node
	std::vector<std::vector<int>> EdgeVarsOf;		// node -> edge list

	void makeNeighLists() {
		HeadNodeOf.resize(npvars);
		EdgeVarsOf.resize(nNodes_);

		for (int i = 0; i < edges.size(); i++) {
			GraphPC::Edge e = edges[i];
			HeadNodeOf[e.cnfvar] = e.to;
			EdgeVarsOf[e.from].push_back(e.cnfvar);
		}
	}

	int start;
	int end;
	int maxpathlen;

	bool allpairs;
	std::vector<int> startnodes;

	// State
	bool unsat;

};

inline size_t Instance::nNodes () const { return nNodes_; }
inline size_t Instance::nEdges () const { return nEdges_; }
inline Glucose::lbool Instance::value (Glucose::Var x) const { return assigns[x]; }
inline Glucose::lbool Instance::value (Glucose::Lit p) const { return assigns[var(p)] ^ sign(p); }

}

#endif
