/*****************************************************************************************[Main.cc]
 *
 * PathCount -- Copyright (c) 2023, Kenji Hashimoto (Nagoya University, Japan)
 *

PathCount sources are based on Glucose 3.0.
We will follow the permissions and copyrights of them.

-----------------------------------------------------------

 Glucose -- Copyright (c) 2009, Gilles Audemard, Laurent Simon
				CRIL - Univ. Artois, France
				LRI  - Univ. Paris Sud, France

Glucose sources are based on MiniSat (see below
 copyrights). Permissions and copyrights of
Glucose are exactly the same as Minisat on which it is based on. (see below).

---------------

Copyright (c) 2003-2006, Niklas Een, Niklas Sorensson
Copyright (c) 2007-2010, Niklas Sorensson

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 **************************************************************************************************/

#include <errno.h>

#include <signal.h>
#include <zlib.h>
#include <gmp.h>
#include <gmpxx.h>
#include <iostream>
#include <fstream>

#include "utils/System.h"
#include "utils/Options.h"

#include "counter/Config.h"
#include "counter/Counter.h"


using namespace Glucose;
using namespace GraphPC;
using namespace std;

//=================================================================================================
static Solver *counter;

// Terminate by notifying the solver and back out gracefully. This is mainly to have a test-case
// for this feature of the Solver as it may take longer than an immediate call to '_exit()'.
static void SIGINT_interrupt(int signum) {
	printf("c o\n");
	printf("c o *** INTERRUPTED by signal %d ***\n", signum);
	printf("c o Elapsed time %.2lf s\n", cpuTime());
	fflush(stdout);
	if(counter != NULL)
		counter->interrupt();
	else
		_exit(1);
}

// Note that '_exit()' rather than 'exit()' has to be used. The reason is that 'exit()' calls
// destructors and may cause deadlocks if a malloc/free function happens to be running (these
// functions are guarded by locks for multithreaded use).
static void SIGINT_exit(int signum) {
	printf("c o\n");
	printf("c o *** INTERRUPTED by signal %d ***\n", signum);
	printf("c o Elapsed time %.2lf s\n", cpuTime());
	fflush(stdout);
	_exit(1); }

static void SetSigAct() {
	signal(SIGINT,  SIGINT_interrupt);  //  2, SIGINT
	// signal(SIGABRT, SIGINT_exit);       //  6, SIGABRT
	signal(SIGSEGV, SIGINT_exit);       // 11, SIGSEGV
	signal(SIGTERM, SIGINT_interrupt);  // 15, SIGTERM
	signal(SIGXCPU, SIGINT_exit);
}
//=================================================================================================
static void printResult(bool finished, bool sat, const mpz_class& result) {
	const char *msg = finished ? "complete" : "incomplete";
	if(!sat) {
		printf("c %s path count 0\n", msg);
	} else {
		gmp_printf("c %s path count %Zd\n", msg, result.get_mpz_t());
	}
}

//=================================================================================================
// Main:
void main_mc(Counter& S, string cnffile, string mapfile)
{
	// Loading input
	cout << "c o Reading from the files ..." << endl;
	std::ifstream in_cnf(cnffile);
	std::ifstream in_map(mapfile);
	if (in_cnf && in_map) {
		S.load(in_cnf, in_map);
		in_cnf.close();
		in_map.close();
	}
	else {
		std::cerr << "Cannot open files." << std::endl;
		exit(1);
	}

	printf("c o Reading finished.\n");
	printf("c o Elapsed time %.2lf s\nc o\n", cpuTime());
	fflush(stdout);

	// Simplifying
	printf("c o Simplification starts...\n");
	bool done = S.preprocess();
	printf("c o Simplification finished.\n");
	printf("c o Elapsed time %.2lf s\nc o\n", cpuTime());
	fflush(stdout);

	if(done) {
		printf("c o Solved by simplification.\n");
		S.printStats();
		printf("c o [Result]\n");
		printResult(true, S.isSat(), S.getCount());
		return;
	}

	// Counting
	counter = (Solver*)&S;
	printf("c o Counting starts ...\n");fflush(stdout);
	bool suc = S.countPaths();
	S.printStats();
	printf("c o [Result]\n");
	printResult(suc, S.isSat(), S.getCount());
	counter = NULL;
	fflush(stdout);
}

int main(int argc, char** argv)
{

	try {
		double start = realTime();

		setUsageHelp("c USAGE: %s [options] <input-CNFfile> <input-GraphMapfile>\n");
		// setUsageHelp("c USAGE: %s [options] <input-file> <result-output-file>\n\n  where input may be either in plain or gzipped DIMACS.\n");
		// printf("This is MiniSat 2.0 beta\n");

#if defined(__linux__)
		fpu_control_t oldcw, newcw;
		_FPU_GETCW(oldcw); newcw = (oldcw & ~_FPU_EXTENDED) | _FPU_DOUBLE; _FPU_SETCW(newcw);
		printf("c o WARNING: for repeatability, setting FPU to use double precision\n");
#endif
		// Extra options:
		//
		IntOption    verb("MAIN", "verb",   "Verbosity level (0=silent, 1=some).", 1, IntRange(0, 1));
		IntOption    cpu_lim("MAIN", "cpu-lim","Limit on CPU time allowed in seconds.\n", INT32_MAX, IntRange(0, INT32_MAX));
		IntOption    mem_lim("MAIN", "mem-lim","Limit on memory usage in megabytes.\n", INT32_MAX, IntRange(0, INT32_MAX));

		parseOptions(argc, argv, true);
		Configuration config;

		if(argc <= 2)
			exit(1);

		string cnffile = argv[1];
		string mapfile = argv[2];

		if(verb) {
			printf("c o "GraphPC_VERSION"\n");
			fflush(stdout);
		}

		// Use signal handlers that forcibly quit until the solver will be able to respond to
		// interrupts:
		SetSigAct();

		// Set limit on CPU-time:
		if (cpu_lim != INT32_MAX){
			rlimit rl;
			getrlimit(RLIMIT_CPU, &rl);
			if (rl.rlim_max == RLIM_INFINITY || (rlim_t)cpu_lim < rl.rlim_max){
				rl.rlim_cur = cpu_lim;
				if (setrlimit(RLIMIT_CPU, &rl) == -1)
					if(verb) printf("c o WARNING! Could not set resource limit: CPU-time.\n");
			} }

		// Set limit on virtual memory:
		if (mem_lim != INT32_MAX){
			rlim_t new_mem_lim = (rlim_t)mem_lim * 1024*1024;
			rlimit rl;
			getrlimit(RLIMIT_AS, &rl);
			if (rl.rlim_max == RLIM_INFINITY || new_mem_lim < rl.rlim_max){
				rl.rlim_cur = new_mem_lim;
				if (setrlimit(RLIMIT_AS, &rl) == -1)
					if(verb) printf("c o WARNING! Could not set resource limit: Virtual memory.\n");
			} }

		Counter S(config);
		main_mc(S, cnffile, mapfile);

		printf("c o \n");
		printf("c o CPU time    = %.2lf s\n", cpuTime());
		printf("c o Real time   = %.2lf s\n", realTime()-start);
		double mem_used = memUsedPeak();
		if (mem_used != 0)
			printf("c o Memory used = %.2f MB\n", mem_used);

		fflush(stdout);
		exit(0);

	} catch (const std::invalid_argument&) {
		printf("c o Invalid argument\n");
		exit(1);
	} catch (OutOfMemoryException&) {
		printf("c o INDETERMINATE (out of memory)\n");
		exit(1);
	} catch(std::bad_alloc& err) {
		printf("c o INDETERMINATE (bad_alloc exception)\n");
		printf("c o Elapsed time %.2lf s\n", cpuTime());
		double mem_used = memUsedPeak();
		if (mem_used != 0)
			printf("c o Memory used = %.2f MB\n", mem_used);
		exit(1);
    }
}
