#ifndef Preprocessor_h
#define Preprocessor_h

#include <vector>
#include <iostream>
#include <fstream>

#include "counter/Config.h"
#include "counter/Instance.h"
#include "gmpxx.h"

#include "preprocessor/Graph.h"

#include "mtl/Vec.h"
#include "utils/System.h"

namespace PPMC {

class Merger {
public:
    std::vector<int> cidx;
    std::vector<std::vector<int>> eqc;
    int num_elem;

    Merger(int nodes) { cidx.resize(nodes, -1); num_elem=0; }

    std::vector<std::vector<int>>& getEquivClasses() { return eqc; }
    int delegateNode(int v) const { return (cidx[v] == -1) ? v : eqc[cidx[v]][0]; }
    int getIndex(int v) const  { return cidx[v]; }

    void merge(int v1, int v2) {
        if(cidx[v1] == -1 && cidx[v2] == -1) {
            cidx[v1] = cidx[v2] = eqc.size();
            eqc.push_back({v1, v2});
            num_elem++;
        }
        else if(cidx[v1] == -1) {
            eqc[cidx[v2]].push_back(v1);
            cidx[v1] = cidx[v2];
        }
        else if(cidx[v2] == -1) {
            eqc[cidx[v1]].push_back(v2);
            cidx[v2] = cidx[v1];
        }
        else {
            assert(cidx[v1] != cidx[v2]);
            MergeEquivClasses(cidx[v1], cidx[v2]);
        }
    }

    void MergeEquivClasses(int c1, int c2) {
        for (int v : eqc[c2]) {
            eqc[c1].push_back(v);
            cidx[v] = c1;
        }
        eqc[c2].clear();
        num_elem--;
    }
};

class Identifier {
public:
	Identifier(int vars) { cidx.resize(2*vars, -1); num_elem = 0; }

	void identify(Glucose::Lit l1, Glucose::Lit l2);

	bool hasEquiv() { return num_elem > 0; }
	std::vector<std::vector<Glucose::Lit>>& getEquivClasses() { return eqc; }
	Glucose::Lit delegateLit(Glucose::Lit l) { return (cidx[toInt(l)] == -1) ? l : eqc[cidx[toInt(l)]][0]; }
	int getIndex(Glucose::Lit l) { return cidx[toInt(l)]; }
	void removeEquivClass(Glucose::Lit l);

private:
	void MergeEquivClasses(int c1, int c2);

	std::vector<std::vector<Glucose::Lit>> eqc;
	std::vector<int> cidx;
	int num_elem;
};

class Preprocessor {
	friend class Solver;
public:
	Preprocessor() : ins(NULL) { };

	void setConfig(GraphPC::ConfigPreprocessor& conf) { this->config = conf; }

	bool Simplify(GraphPC::Instance* ins);

private:
	bool SAT_FLE(bool satsolving);
	bool Strengthen();
	bool MergeAdjEquivs();
	bool VariableEliminate(bool dve);
	void pickVars(std::vector<int>& vars);
	// void pickDefVars(std::vector<int>& vars);
	int ElimVars(const std::vector<Glucose::Var>& del);

	void Compact(const Glucose::vec<Glucose::lbool>& assigns, const std::vector<Glucose::Var>& elimvars={});
	void CompactClauses(const Glucose::vec<Glucose::lbool>& assigns, std::vector<std::vector<Glucose::Lit>>& cls, std::vector<bool>& occurred, int& varnum);
	void RewriteClauses(std::vector<std::vector<Glucose::Lit>>& cls, const std::vector<Glucose::Var>& map);
	void RewriteClauses(std::vector<std::vector<Glucose::Lit>>& cls, const std::vector<Glucose::Lit>& map);
	void Subsume();

	bool isVECandidate(Graph& G, std::vector<int>& freq, int i) const;

	void compactGraph();

	void printCNFInfo(const char* ppname = "", bool initial = false);

	GraphPC::Instance *ins;
	GraphPC::ConfigPreprocessor config;
};

inline void Preprocessor::printCNFInfo(const char* ppname, bool initial)
{
	if(initial)
		printf("c o [%-7s] %d vars (%d pvars), %d cls\n", ppname, ins->vars, ins->npvars, ins->clauses.size());
	else
		printf("c o [%-7s] %d vars (%d pvars), %d cls, %d lrnts, %d fvars, elap. %.2lf s\n", ppname, ins->vars, ins->npvars, ins->clauses.size(), ins->learnts.size(), ins->freevars, Glucose::cpuTime());

	fflush(stdout);
}

}

#endif /* Preprocessor_h */
