#ifndef _CACHE_H_
#define _CACHE_H_

#include <cstdlib>
#include <functional>
#include <unordered_map>
#include <vector>
#include <assert.h>
#include <gmpxx.h>
#include "mtl/Alloc.h"
#include "mtl/Vec.h"

#define WORDSIZE    (sizeof(unsigned int)*8)

using namespace std;

class BitVec {
public:
	// constructor
	BitVec() : _len(0), _data(nullptr) {
	}

	BitVec(int len) {
		_len  = len;
		if(_len > 0) {
			_data = (unsigned int*)malloc(units() * sizeof(unsigned int));
			if(_data == nullptr) {
				printf("c o Malloc Failed : cachekey\n");
				// throw OutOfMemoryException();
				abort();
			}
		}
		else
			_data = nullptr;
	}

	// copy constructor
	BitVec(const BitVec &obj) {
		_len = obj.len();
		// _data = (unsigned int*)calloc(obj.units(), sizeof(unsigned int));
		if(_len > 0) {
			_data = (unsigned int*)malloc(obj.units() * sizeof(unsigned int));
			if(_data == nullptr) {
				printf("c o Malloc Failed : cachekey\n");
				//throw OutOfMemoryException();
				abort();
			}
		}
		else
			_data = nullptr;

		for(int i=0; i<obj.units();i++)
			_data[i] = (obj.getVec())[i];
	}

	~BitVec() {
		if(_data != nullptr) free(_data);
	}

	void setLen(int len) {
		_len = len;
		if(_data != nullptr) {
			free(_data);
			_data = nullptr;
		}
		// _data = (unsigned int*)calloc(units(), sizeof(unsigned int));
		if(_len > 0) {
			_data = (unsigned int*)malloc(units() * sizeof(unsigned int));
			if(_data == nullptr) {
				printf("c o Malloc Failed : cachekey\n");
				// throw OutOfMemoryException();
				abort();
			}
		}
		else
			_data = nullptr;
	}

	unsigned int *getVec() const { return _data; }
	void setVec(const vector<bool> &vec) {
		assert(vec.size() == _len);
		int i=0, u=0;
		while(i<_len) {
			_data[u] = 0;
			for(int p=0; p<WORDSIZE && i<_len; p++, i++) {
				_data[u] |= (vec[i] << p);
			}
			u++;
		}
	}
	int len()   const { return _len; }
	int units() const { return (_len / WORDSIZE) + (_len % WORDSIZE == 0 ? 0 : 1); }

	bool operator==(const BitVec &left) const {
		if(_len != left.len()) return false;

		for(int i=0; i < units(); i++)
			if(_data[i] != (left.getVec())[i])
				return false;

		return true;
	}

	void reset() {
		for(int i=0; i<units(); i++)
			_data[i] = 0;
	}
	// For debug
	void print()
	{
		for(int i=0; i<units(); i++)
			printf("%08x ", _data[i]);
		printf("\n");
	}

private:
	int _len;
	unsigned int *_data;
};

template<typename T>
void hash_combine(size_t & seed, T const& v) {
	seed ^= std::hash<T>{}(v) + 0x9e3779b9 + (seed << 6) + (seed >> 2);
}

template <>
class std::hash<BitVec>{
public:
	size_t operator () (const BitVec &p) const {
		unsigned int *d = p.getVec();

		size_t seed = 0;
		for(int i = 0; i < p.units(); i++)
			hash_combine(seed, d[i]);

		seed += p.units();
		return seed;
	}
};

class Cache {
	vector<unordered_map<BitVec, vector<mpz_class>>> _map;

	uint64_t capacity;	// in bytes (default: 4GB)
	uint64_t size;		// in bytes
	int keysize;			// in bytes
	int elemsize;			// in bytes

	uint64_t searches, hits;
	vector<int> searchesOf;
	vector<int> hitsOf;

	int nreduce;

public:
	Cache() : capacity(4294967296), size(0), keysize(0), elemsize(0), searches(0), hits(0),nreduce(0) {}
	~Cache() {}

	void setCapacity(int cap) {
		// capacity = cap MB  (in bytes)
		capacity = ((uint64_t) cap) * 1048576;
	}

	void resize(int num) {
		_map.resize(num);
		searchesOf.resize(num);
		hitsOf.resize(num);

		on.clear();
		on.growTo(num, true);

		keysize  = ((num / WORDSIZE) + (num % WORDSIZE == 0 ? 0 : 1)) * (WORDSIZE/8);
		elemsize = keysize + sizeof(std::pair<BitVec, std::vector<mpz_class>>);
	}

	void insert(int v, BitVec &k, const vector<mpz_class>& val) {
		_map[v].emplace(k, val);

		size += elemsize + sizeof(std::vector<mpz_class>)*val.size();
		if(size > capacity) {
			clear();

			for(int i=0; i<_map.size(); i++) {
				if(on[i] && (hitsOf[i] == 0)) {
					on[i] = false;
					unordered_map<BitVec, std::vector<mpz_class>>().swap(_map[i]);
				}
			}

			// ToDo: think a better cache reduction strategy.
		}
	}

	void clear() {
		for(int i=0; i<_map.size(); i++) {
			_map[i].clear();
			// unordered_map<BitVec, opt_obdd_t*>().swap(_map[i]);
		}
		size = 0;
		// printf("cache reduced\n");
		nreduce++;
	}

	bool find(int v, const BitVec &k, vector<mpz_class> &result) {
		// If k is hit, update node by the result and return true.
		// Otherwise, result is "not updated" and return false.

		searches++; searchesOf[v]++;
		auto itr = _map[v].find(k);
		if(itr != _map[v].end()) {
			hits++; hitsOf[v]++;
			result = itr->second;
			return true;
		}
		else
			return false;
	}

	void printStats() const {
		printf("c o cacheHitRate          = %-6.2f %%    (%"PRIu64" searches, %"PRIu64" hits)\n", (double)hits*100/searches, searches, hits);
		printf("c o last cacheSize        = %-6.2f MB   (rough estimate)\n", (double)size/1048576);
		printf("c o #cacheReductions      = %-11d (capacity lim: around %"PRIu64" MB)\n", nreduce, capacity/(1024*1024));

		//for(int i=0; i<_map.size(); i++)
		//   printf("c [%3d] searches %12d, hits %12d\n", i, searchesOf[i], hitsOf[i]);
	}

	Glucose::vec<bool> on;
};

#endif /* _CACHE_H_ */
