#include <gmpxx.h>
#include "counter/Counter.h"
#include "mtl/Sort.h"
#include "utils/System.h"

#include <cstdio>
#include <cstdlib>
#include <fstream>
#include <iostream>

using namespace Glucose;
using namespace GraphPC;

//=================================================================================================
// Constructor

Counter::Counter(Configuration& config_):
  limlevel           (0)
, fixed_level_		(0)
, on_bj				(config_.cntr.backjump)
, on_simp				(config_.cntr.remove_sat_cls)
, on_restart			(config_.cntr.restart)
, progress				(INIT)
, setStart				(false)
, startvar				(var_Undef)
, sat					(false)
, npaths           	(0)
, config				(config_.cntr)
, verbosity_c        (1)
, npvars_isolated    (0)
, nbackjumps         (0)
, nbackjumps_sp      (0)
, nrestarts          (0)
, reduce_dbs_pre     (0)
, simp_dbs           (0)
, simplify_time      (0.0)
{
	verbosity = 0;
	showModel = false;

	pp.setConfig(config_.pp);
}
//=================================================================================================
// Load instance
void Counter::load(std::istream& cnf, std::istream& map)
{
	ins.loadCNF(cnf);
	ins.loadGraph(map, config.allpairs);

	ins.printGraphStats("original");
	progress = LOADED;
}
//=================================================================================================
// Preprocessing
bool Counter::preprocess()
{
	/*
	if(!config.preprocess) {
		printf("c o skip preprocessing\n");
		return false;
	}
	*/
	sat = pp.Simplify(&ins);

	bool done = ins.unsat || ins.vars == 0;

	if(ins.unsat) {
		npaths.clear();
		progress = COMPLETED_BYPP;
	}
	else if (ins.vars == 0) {
		// npaths = 1 << ins.freevars;
		npaths.resize(ins.maxpathlen_orig-ins.maxpathlen, 0);
		npaths.push_back(1);
		npvars_isolated = ins.freevars;
		progress = COMPLETED_BYPP;
	}
	else {
		progress = PREPROCESSED;
		if (config.pp_outfile != "NULL") {
			printf("c o outputing simplified CNF...");
			std::ofstream out(config.pp_outfile);
			if (out) {
				ins.toDimacs(out);
				out.close();
				printf("done\n");
			}
			else {
				std::cerr << "Cannot open file:" << config.pp_outfile << std::endl;
			}
		}

		if (config.graph_outfile != "NULL") {
			printf("c o outputing simplified graph...");
			std::ofstream out(config.graph_outfile);
			if (out) {
				ins.writeGraph(out);
				out.close();
				printf("done\n");
			}
			else {
				std::cerr << "Cannot open file:" << config.graph_outfile << std::endl;
			}
		}

		// ins.printVarMapStats();
	}
	ins.printGraphStats("simplied");
	return done;
}

// Inprocessing
bool Counter::simplify()
{
	if (!ok || propagate() != CRef_Undef)
		return ok = false;

	if (nAssigns() == simpDB_assigns || (simpDB_props > 0))
		return true;

	double cpu_time = cpuTime();

	// Remove satisfied clauses:
	removeSatisfied(learnts);
	if (remove_satisfied)        // Can be turned off.
		removeSatisfied(clauses);
	// checkGarbage();
	// rebuildOrderHeap();

	simpDB_assigns = nAssigns();
	simpDB_props   = clauses_literals + learnts_literals;   // (shouldn't depend on stats really, but it will do for now)

	simp_dbs++;
	simplify_time += cpuTime() - cpu_time;

	if(simp_dbs >= config.rmvsatcl_threshold)
		on_simp = false;
	return true;
}

//=================================================================================================
// Major methods:
bool Counter::countPaths()
{
	if (progress != PREPROCESSED) {
		printf("c o preprocessing is not done yet, or model counting is already done.\n");
		return false;
	}

	ins.makeNeighLists();
	import();

	solves = starts = 0;
	// if(config.allpairs)
	// if(ins.allpairs)
		// count_main_all();
	// else
		count_main(ins.start, ins.end);

	sat = topDecision().hasModel();
	assert(ins.freevars == 0);
	npaths = topDecision().totalModels(); // << ins.freevars;
	cancelUntil(0);

	if(ins.maxpathlen_orig-ins.maxpathlen>0) {
		std::vector<mpz_class> tmp(ins.maxpathlen_orig-ins.maxpathlen, 0);
		tmp.insert(tmp.end(),npaths.begin(), npaths.end());
		npaths = tmp;
	}
//	assert(npaths.size() <= ins.maxpathlen_orig+1);
//	npaths.resize(ins.maxpathlen_orig+1, 0);

	progress = asynch_interrupt ? FAILED : COMPLETED;
	return !asynch_interrupt;
}

void Counter::count_main(int startnode, int endnode)
{
	int          backtrack_level;
	int          conflictC = 0;
	vec<Lit>     learnt_clause,selectors;
	unsigned int nblevels,szWoutSelectors;
	bool blocked=false;

	btStateT bstate = RESOLVED;
	limlevel = 0;

	nodeTrail.capacity(ins.maxpathlen+1);
	nodeTrail.push(startnode);
	initDecision();
	topDecision().changeBranch();

	initCache();
	key.setLen(ins.EdgeVarsOf.size());

	for(;;) {
		if(asynch_interrupt){
			printf("c o !!! Interrupted !!!\n");
			while(bstate != EXIT)
				bstate = backtrack();
		}
		if(bstate == EXIT) break;

		for(int i = 0; i < unitcls.size(); i++){
			enqueue(unitcls[i]);
			vardata[var(unitcls[i])].level = 0;
			activity[var(unitcls[i])] = 0;
		}
		if(decisionLevel() == 0) unitcls.clear();

		CRef confl = propagate();

		if(confl != CRef_Undef){
			// CONFLICT
			conflicts++;conflictC++;conflictsRestarts++;
			if(conflicts % 5000 == 0 && var_decay < 0.95)
				var_decay += 0.01;

			if(decisionLevel() == 0) break;

			trailQueue.push(trail.size());
			// BLOCK RESTART (CP 2012 paper)
			if( conflictsRestarts>LOWER_BOUND_FOR_BLOCKING_RESTART && lbdQueue.isvalid()  && trail.size()>R*trailQueue.getavg()) {
				lbdQueue.fastclear();
				nbstopsrestarts++;
				if(!blocked) {lastblockatrestart=starts;nbstopsrestartssame++;blocked=true;}
			}

			learnt_clause.clear();
			selectors.clear();
			bool suc = analyzeMC(confl, learnt_clause, selectors, backtrack_level, nblevels, szWoutSelectors);
			if(!suc) {
				nbackjumps_sp++;
				bjResolve(backtrack_level);
				if(backtrack_level < limlevel) limlevel = backtrack_level;
				bstate = RESOLVED;
				continue;
			}

			lbdQueue.push(nblevels);
			sumLBD += nblevels;

			CRef cr = CRef_Undef;
			if (learnt_clause.size() == 1){
				unitcls.push(learnt_clause[0]);
				nbUn++;
			}
			else {
				cr = ca.alloc(learnt_clause, true);
				ca[cr].setLBD(nblevels); 
				ca[cr].setSizeWithoutSelectors(szWoutSelectors);
				if(nblevels<=2) nbDL2++;
				if(ca[cr].size()==2) nbBin++;

				learnts.push(cr);
				attachClause(cr);
				claBumpActivity(ca[cr]);
			}

			varDecayActivity();
			claDecayActivity();

			bstate = Resolve(backtrack_level, cr, learnt_clause[0]);
		} else {
			// NO CONFLICT

			if (on_restart && (lbdQueue.isvalid() && ((lbdQueue.getavg()*K) > (sumLBD / conflictsRestarts)))) {
				lbdQueue.fastclear();
				nrestarts++;
				// printf("c o backjump from dl %d to fixed_level_ %d\n", decisionLevel(), fixed_level_);
				bjResolve(fixed_level_);
				limlevel = fixed_level_;
				bstate = RESOLVED;
				continue;
			}

			if (on_simp && checkfixedDL() && !simplify()) {
				topDecision().increaseModels(0, false);
				bstate = backtrack();
				if(bstate == EXIT) break;
				else { assert(false); abort(); }
			}

			if(conflicts>=(uint64_t)curRestart* nbclausesbeforereduce)
			{
				assert(learnts.size()>0);
				curRestart = (conflicts/ nbclausesbeforereduce)+1;
				reduceDB();
				nbclausesbeforereduce += incReduceDB;
			}

			// DECIDE AN EDGE
			Var next_edge = findNextEdge();
			topDecision().setLen(nodeTrail.size()-topDecision().basePos());

			if (next_edge == var_Undef) {
				assert(nodeTrail.last() == endnode);

				bool dirty = false;
				int current_dl = decisionLevel();

				newDecisionLevel();
				for(int i=0; i<ins.npvars; i++) {
					if(value(i) == l_Undef) {
						uncheckedEnqueue(~mkLit(i));

						if(propagate() != CRef_Undef) {
							if(current_dl <= 3) {
								conflicts++;

								learnt_clause.clear();
								for(int i=current_dl-1; i>=0;i--)
									learnt_clause.push(~trail[trail_lim[i]]);

								lbdQueue.push(learnt_clause.size());
								sumLBD += learnt_clause.size();

								if (learnt_clause.size() == 1) {
									unitcls.push(learnt_clause[0]);
									nbUn++;
								}
								else {
									CRef cr = ca.alloc(learnt_clause, true);
									ca[cr].setLBD(learnt_clause.size());
									if(nblevels<=2) nbDL2++;
									if(ca[cr].size()==2) nbBin++;

									learnts.push(cr);
									attachClause(cr);
									claBumpActivity(ca[cr]);
								}
								varDecayActivity();
								claDecayActivity();
							}

							dirty = true;
							break;
						}
					}
				}
				cancelUntil(current_dl);

				if(dirty)
					topDecision().increaseModels(0, false);
				else
					topDecision().increaseModels(1);
				bstate = backtrack();
			}
			else {
				if (config.cache && cache.on[nodeTrail.last()]) {
					genkey(key);
					// key.print();
					vector<mpz_class> res;
					if (cache.find(nodeTrail.last(), key, res)) {
						topDecision().increaseModels(res);
						bstate = backtrack();
						continue;
					}
				}

				decisions++;

				nodeTrail.push(ins.HeadNodeOf[next_edge]);
				newDecisionLevel();
				pushDecision(trail.size());

				Lit dlit = mkLit(next_edge);	// make it true first
				uncheckedEnqueue(dlit);
			}
		}
	}
}
/*
void Counter::count_main_all()
{
	int          backtrack_level;
	int          conflictC = 0;
	vec<Lit>     learnt_clause,selectors;
	unsigned int nblevels,szWoutSelectors;
	bool blocked=false;

	int nextStart = 0;

	btStateT bstate = RESOLVED;
	limlevel = 0;

	nodeTrail.capacity(ins.maxpathlen+1);
	pushDecision(0);
	topDecision().changeBranch();

	setStart=false;

	initCache();
	key.setLen(ins.nNodes());

	for(;;) {
		if(asynch_interrupt){
			printf("c o !!! Interrupted (%d/%d)!!!\n", nextStart, ins.nNodes());
			while(bstate != EXIT)
				bstate = backtrack();
		}
		if(bstate == EXIT) break;

		for(int i = 0; i < unitcls.size(); i++){
			enqueue(unitcls[i]);
			vardata[var(unitcls[i])].level = 0;
			activity[var(unitcls[i])] = 0;
		}
		if(decisionLevel() == 0) unitcls.clear();

		CRef confl = propagate();

		if(confl != CRef_Undef){
			// CONFLICT
			conflicts++;conflictC++;conflictsRestarts++;
			if(conflicts % 5000 == 0 && var_decay < 0.95)
				var_decay += 0.01;

			if(decisionLevel() == 0) break;

			trailQueue.push(trail.size());
			// BLOCK RESTART (CP 2012 paper)
			if( conflictsRestarts>LOWER_BOUND_FOR_BLOCKING_RESTART && lbdQueue.isvalid()  && trail.size()>R*trailQueue.getavg()) {
				lbdQueue.fastclear();
				nbstopsrestarts++;
				if(!blocked) {lastblockatrestart=starts;nbstopsrestartssame++;blocked=true;}
			}

			learnt_clause.clear();
			selectors.clear();
			bool suc = analyzeMC(confl, learnt_clause, selectors, backtrack_level, nblevels, szWoutSelectors);
			if(!suc) {
				nbackjumps_sp++;
				bjResolve(backtrack_level);
				if(backtrack_level < limlevel) limlevel = backtrack_level;
				bstate = RESOLVED;
				continue;
			}

			lbdQueue.push(nblevels);
			sumLBD += nblevels;

			CRef cr = CRef_Undef;
			if (learnt_clause.size() == 1){
				unitcls.push(learnt_clause[0]);
				nbUn++;
			}
			else {
				cr = ca.alloc(learnt_clause, true);
				ca[cr].setLBD(nblevels);
				ca[cr].setSizeWithoutSelectors(szWoutSelectors);
				if(nblevels<=2) nbDL2++;
				if(ca[cr].size()==2) nbBin++;

				learnts.push(cr);
				attachClause(cr);
				claBumpActivity(ca[cr]);
			}

			varDecayActivity();
			claDecayActivity();

			bstate = Resolve(backtrack_level, cr, learnt_clause[0]);
		} else {
			// NO CONFLICT

			if (on_restart && (lbdQueue.isvalid() && ((lbdQueue.getavg()*K) > (sumLBD / conflictsRestarts)))) {
				lbdQueue.fastclear();
				nrestarts++;
				// printf("c o backjump from dl %d to fixed_level_ %d\n", decisionLevel(), fixed_level_);
				bjResolve(fixed_level_);
				limlevel = fixed_level_;
				bstate = RESOLVED;
				continue;
			}

			if (on_simp && checkfixedDL() && !simplify()) {
				topDecision().increaseModels(0, false);
				bstate = backtrack();
				if(bstate == EXIT) break;
				else { assert(false); abort(); }
			}

			if(conflicts>=(uint64_t)curRestart* nbclausesbeforereduce)
			{
				assert(learnts.size()>0);
				curRestart = (conflicts/ nbclausesbeforereduce)+1;
				reduceDB();
				nbclausesbeforereduce += incReduceDB;
			}

			Var next = var_Undef;
			if(!setStart) {
				// DECIDE A START NODE
				for(int i=nextStart; i<ins.startnodes.size(); i++) {
					Var v = ins.startnodes[i];
					if(v == var_Undef) continue;

					if(value(v) != l_False) {
						nodeTrail.push(i);
						if(value(v) == l_Undef) {
							decisions++;
							newDecisionLevel();
							pushDecision(trail.size());
							Lit dlit = mkLit(v);
							uncheckedEnqueue(dlit);
							startvar = v;
						}
						setStart = true;
						nextStart = i;
						if(config.cache)
							cache.clear();
						limlevel = decisionLevel();
						break;
					}
				}
				if(config.progressbar) {
					double percent = (double)nextStart/ins.nNodes() * 100;
					double elapsed = cpuTime();
					double est_remaining = percent > 0 ? elapsed/percent*100-elapsed : -1;
					int round   = (int) (percent / 4);
					printf("\033[1F\033[G\033[2K");
					printf("c o %5.1f%% [%s>%s] (%d/%d), elaspsed: %5.1f s, remaining: %5.1f s\n", percent, std::string(round, '=').c_str(), std::string(24-round, ' ').c_str(), nextStart, ins.nNodes(), elapsed, est_remaining);
					fflush(stdout);
				}
				continue;
			} else {
				// DECIDE AN EDGE
				Var next_edge = findNextEdge();

				if (next_edge == var_Undef) {
					topDecision().increaseModels(1);
					bstate = backtrack();
				}
				else {
					if (config.cache && cache.on[nodeTrail.last()]) {
						genkey(key);
						// key.print();
						mpz_class res;
						if (cache.find(nodeTrail.last(), key, res)) {
							topDecision().increaseModels(res);
							bstate = backtrack();
							continue;
						}
					}

					decisions++;

					nodeTrail.push(ins.HeadNodeOf[next_edge]);
					newDecisionLevel();
					pushDecision(trail.size());

					Lit dlit = mkLit(next_edge);	// make it true first
					uncheckedEnqueue(dlit);
				}
			}
		}
	}
	if(config.progressbar && !asynch_interrupt) {
		printf("\033[1F\033[G\033[2K");
		printf("c o 100%% [%s>] (%d/%d)\n", std::string(24, '=').c_str(), ins.nNodes(), ins.nNodes());
		fflush(stdout);
	}
}
*/

Var Counter::findNextEdge()
{
	bool proceed = true;

	Var dec_edge = var_Undef;
	double maxscore = -1.0;
	int v_cur = nodeTrail.last();

	while(proceed && nodeTrail.size() <= ins.maxpathlen) {
		proceed = false;

		for(int i=0; i<ins.EdgeVarsOf[v_cur].size(); i++) {
			Var edge = ins.EdgeVarsOf[v_cur][i];
			if(value(edge) == l_Undef) {
				if(activity[edge] > maxscore) {
					dec_edge = edge;
					maxscore = activity[edge];
					if(i > 0)
						std::swap(ins.EdgeVarsOf[v_cur][i], ins.EdgeVarsOf[v_cur][0]);
				}
			}
			else if(value(edge) == l_True) {
				if(i > 0)
					std::swap(ins.EdgeVarsOf[v_cur][i], ins.EdgeVarsOf[v_cur][0]);
				v_cur = ins.HeadNodeOf[edge];
				nodeTrail.push(v_cur);
				proceed = true;
				dec_edge = var_Undef;
				break;
			}
		}
	}

	return dec_edge;
}

bool Counter::analyzeMC(CRef confl, vec<Lit>& out_learnt, vec<Lit>&selectors,
		int& out_btlevel, unsigned int &lbd, unsigned int &szWithoutSelectors) {
	int pathC = 0;
	Lit p = lit_Undef;

	// Generate conflict clause:
	//
	out_learnt.push();      // (leave room for the asserting literal)
	int index = trail.size() - 1;

	do {
		assert(confl != CRef_Undef); // (otherwise should be UIP)
		Clause& c = ca[confl];

		// Special case for binary clauses
		// The first one has to be SAT
		if (p != lit_Undef && c.size() == 2 && value(c[0]) == l_False) {

			assert(value(c[1])==l_True);
			Lit tmp = c[0];
			c[0] = c[1], c[1] = tmp;
		}

		if (c.learnt())
			claBumpActivity(c);

#ifdef DYNAMICNBLEVEL
		// DYNAMIC NBLEVEL trick (see competition'09 companion paper)
		if (c.learnt() && c.lbd() > 2) {
			unsigned int nblevels = computeLBDMC(c);
			if (nblevels + 1 < c.lbd()) { // improve the LBD
				if (c.lbd() <= lbLBDFrozenClause) {
					c.setCanBeDel(false);
				}
				// seems to be interesting : keep it for the next round
				c.setLBD(nblevels); // Update it
			}
		}
#endif

		for (int j = (p == lit_Undef) ? 0 : 1; j < c.size(); j++) {
			Lit q = c[j];

			if (!seen[var(q)] && level(var(q)) > 0) {
				if (!isSelector(var(q)))
					varBumpActivity(var(q));
				seen[var(q)] = 1;
				if (level(var(q)) >= decisionLevel()) {
					pathC++;
#ifdef UPDATEVARACTIVITY
					// UPDATEVARACTIVITY trick (see competition'09 companion paper)
					if (!isSelector(var(q)) && (reason(var(q)) != CRef_Undef)
							&& ca[reason(var(q))].learnt())
						lastDecisionLevel.push(q);
#endif

				} else {
					if (isSelector(var(q))) {
						assert(value(q) == l_False);
						selectors.push(q);
					} else
						out_learnt.push(q);
				}
			}
		}

		// added by k-hasimt
		if (pathC <= 0) {
			// backjump conservatively to restart
			out_btlevel = 0;
			for (int i = 1; i < out_learnt.size(); i++) {
				seen[var(out_learnt[i])] = 0;
				if (out_btlevel < level(var(out_learnt[i])))
					out_btlevel = level(var(out_learnt[i]));
			}
			if (out_btlevel > 0)
				out_btlevel--;
			for (int j = 0; j < selectors.size(); j++)
				seen[var(selectors[j])] = 0;
			// ToDo: make an appropriate learnt clause and do backjumping.
			return false;
		}
		//---
		/*
		 // Select next clause to look at:
		 while (!seen[var(trail[index--])]);
		 p     = trail[index+1];
		 confl = reason(var(p));
		 seen[var(p)] = 0;
		 pathC--;
		 */
		do {
			// Select next clause to look at:
			while (!seen[var(trail[index--])])
				;
			p = trail[index + 1];
			confl = reason(var(p));
			seen[var(p)] = 0;
			pathC--;
			if (confl == CRef_Undef && pathC > 0) {
				out_learnt.push(~p);
				assert(level(var(p)) <= decisionLevel());
			}
		} while (confl == CRef_Undef && pathC > 0);

	} while (pathC > 0);
	out_learnt[0] = ~p;

	// Simplify conflict clause:
	//
	int i, j;

	for (int i = 0; i < selectors.size(); i++)
		out_learnt.push(selectors[i]);

	out_learnt.copyTo(analyze_toclear);
	if (ccmin_mode == 2) {
		uint32_t abstract_level = 0;
		for (i = 1; i < out_learnt.size(); i++)
			abstract_level |= abstractLevel(var(out_learnt[i])); // (maintain an abstraction of levels involved in conflict)

		for (i = j = 1; i < out_learnt.size(); i++)
			if (reason(var(out_learnt[i])) == CRef_Undef
					|| !litRedundant(out_learnt[i], abstract_level))
				out_learnt[j++] = out_learnt[i];

	} else if (ccmin_mode == 1) {
		for (i = j = 1; i < out_learnt.size(); i++) {
			Var x = var(out_learnt[i]);

			if (reason(x) == CRef_Undef)
				out_learnt[j++] = out_learnt[i];
			else {
				Clause& c = ca[reason(var(out_learnt[i]))];
				// Thanks to Siert Wieringa for this bug fix!
				for (int k = ((c.size() == 2) ? 0 : 1); k < c.size(); k++)
					if (!seen[var(c[k])] && level(var(c[k])) > 0) {
						out_learnt[j++] = out_learnt[i];
						break;
					}
			}
		}
	} else
		i = j = out_learnt.size();

	max_literals += out_learnt.size();
	out_learnt.shrink(i - j);
	tot_literals += out_learnt.size();

	/* ***************************************
	 Minimisation with binary clauses of the asserting clause
	 First of all : we look for small clauses
	 Then, we reduce clauses with small LBD.
	 Otherwise, this can be useless
	 */
	if (!incremental && out_learnt.size() <= lbSizeMinimizingClause) {
		minimisationWithBinaryResolution(out_learnt);
	}
	// Find correct backtrack level:
	//
	if (out_learnt.size() == 1)
		out_btlevel = 0;
	else {
		int max_i = 1;
		// Find the first literal assigned at the next-highest level:
		for (int i = 2; i < out_learnt.size(); i++)
			if (level(var(out_learnt[i])) > level(var(out_learnt[max_i])))
				max_i = i;
		// Swap-in this literal at index 1:
		Lit p = out_learnt[max_i];
		out_learnt[max_i] = out_learnt[1];
		out_learnt[1] = p;
		out_btlevel = level(var(p));
	}

	// Compute the size of the clause without selectors (incremental mode)
	if (incremental) {
		szWithoutSelectors = 0;
		for (int i = 0; i < out_learnt.size(); i++) {
			if (!isSelector(var((out_learnt[i]))))
				szWithoutSelectors++;
			else if (i > 0)
				break;
		}
	} else
		szWithoutSelectors = out_learnt.size();

	// Compute LBD
	lbd = computeLBDMC(out_learnt, out_learnt.size() - selectors.size());

#ifdef UPDATEVARACTIVITY
	// UPDATEVARACTIVITY trick (see competition'09 companion paper)
	if (lastDecisionLevel.size() > 0) {
		for (int i = 0; i < lastDecisionLevel.size(); i++) {
			if (ca[reason(var(lastDecisionLevel[i]))].lbd() < lbd)
				varBumpActivity(var(lastDecisionLevel[i]));
		}
		lastDecisionLevel.clear();
	}
#endif

	for (int j = 0; j < analyze_toclear.size(); j++)
		seen[var(analyze_toclear[j])] = 0;    // ('seen[]' is now cleared)
	for (int j = 0; j < selectors.size(); j++)
		seen[var(selectors[j])] = 0;

	return true;
}

inline void Counter::bjResolve(int level) {
	cancelUntil(level);

	while (dl_.size() > (unsigned) level + 1)
		popDecision();

	topDecision().increaseModels(0, false);	// reset
}

inline btStateT Counter::Resolve(int bk_level, CRef cr, Lit lit) {
	int level = std::max(bk_level, limlevel);
	bool canjump = on_bj && level+1 < decisionLevel() && (double)level/decisionLevel() < config.bj_threshold;

	if(canjump) {
		nbackjumps++;
		bjResolve(level);
		if(cr != CRef_Undef)
			uncheckedEnqueue(lit, cr);
		if(nbackjumps > ins.nEdges()) {
			on_bj = false;
		}
		return RESOLVED;
	}
	else {
		topDecision().increaseModels(0, false);	// set 0 (no model found at the current branch)
		return backtrack();
	}
}

btStateT Counter::backtrack() {
	if (decisionLevel() == 0) return EXIT;

	for (;;) {
		assert(decisionLevel() > 0);

		if (topDecision().isFirstBranch()) {
			// FirstBranch

			Lit dlit = trail[trail_lim.last()];
			cancelCurDL();
			nodeTrail.shrink(nodeTrail.size()-topDecision().basePos());
			uncheckedEnqueue(~dlit);
			// Note: we do not try to assign an extra lit by the learnt clause when conflict occurs, but it may be an option.

			if(topDecision().hasModel()) limlevel = decisionLevel();
			topDecision().changeBranch();

			if(setStart && var(dlit) == startvar)
				setStart = false;

			return RESOLVED;
		} else {
			// SecondBranch

			vector<mpz_class> total = topDecision().totalModels();
			bool hasmodel = topDecision().hasModel();
			prevDecision().increaseModels(total, hasmodel);

			cancelCurDL();
			trail_lim.pop();
			popDecision();

			// assert(hasmodel || total == 0);

			if(config.cache && hasmodel && topDecision().isFirstBranch() && cache.on[nodeTrail.last()]) {
			// if(config.cache && topDecision().isFirstBranch() && cache.on[nodeTrail.last()]) {
				// Note: If topDecision()'s branch is second, the current decision chooses no edge.
				//       To avoid duplicate keys generated from nodeTrail, only when the branch is first, the result is inserted to the cache.

				genkey(key);
				cache.insert(nodeTrail.last(), key, total);
			}

			if (decisionLevel() == 0)
				return EXIT;
		}
	}

	assert(false);
	return EXIT;
}

void Counter::cancelCurDL() {
	for (int c = trail.size() - 1; c >= trail_lim.last(); c--) {
		Var x = var(trail[c]);
		assigns[x] = l_Undef;
		polarity[x] = sign(trail[c]);
	}
	qhead = trail_lim.last();
	trail.shrink(trail.size() - trail_lim.last());
}

//=================================================================================================
// Print methods:
void Counter::printStats() const
{
	if(verbosity_c) {
		if(progress == COMPLETED || progress == FAILED) {
			printf("c o [Statistics]\n");
			printf("c o conflicts             = %-11"PRIu64"\n", conflicts);
			printf("c o decisions             = %-11"PRIu64"\n", decisions);
			printf("c o propagations          = %-11"PRIu64"\n", propagations);
			printf("c o simp dbs              = %-11"PRIu64" (%.3f s)\n", simp_dbs, simplify_time);
			printf("c o reduce dbs            = %-11"PRIu64"\n", nbReduceDB);
			printf("c o learnts (uni/bin/lbd2)= %"PRIu64"/%"PRIu64"/%"PRIu64"\n", nbUn, nbBin, nbDL2);
			printf("c o last learnts          = %-11d (%"PRIu64" learnts removed, %4.2f%%)\n", learnts.size(), nbRemovedClauses, nbRemovedClauses * 100 / (double)conflicts);

			printf("c o backjumps             = %-11"PRIu64" (sp %"PRIu64")\n", nbackjumps+nbackjumps_sp, nbackjumps_sp);
			printf("c o restarts              = %-11"PRIu64"\n", nrestarts);
			if(config.cache)
				cache.printStats();
			printf("c o\n");
		}

		fflush(stdout);
	}
}


//=================================================================================================
// Methods for Importing instance and additional information

void Counter::import()
{
	int nvars = ins.vars;
	watches  .init(mkLit(nvars-1, true ));
	watchesBin  .init(mkLit(nvars-1, true ));
	assigns  .growTo(nvars, l_Undef);
	vardata  .growTo(nvars, mkVarData(CRef_Undef, 0));
	activity .growTo(nvars, 0);
	seen     .growTo(nvars, 0);
	permDiff  .growTo(nvars, 0);
	polarity .growTo(nvars, true);
	decision .growTo(nvars);
	trail    .capacity(nvars);
	for(int i=0; i<nvars; i++)
		setDecisionVar(i, true);

	for(int i=0; i<ins.assignedLits.size(); i++) {
		uncheckedEnqueue(ins.assignedLits[i]);
	}

	CRef cr;
	for(auto c : ins.clauses) {
		cr = ca.alloc(c, false);
		clauses.push(cr);
		attachClause(cr);
	}
	for(auto c : ins.learnts) {
		cr = ca.alloc(c, true);
		clauses.push(cr);
		attachClause(cr);
	}

	sat = !ins.unsat;
}

//=================================================================================================
// Methods for Cache
void Counter::initCache()
{
	cache.resize(ins.nNodes());
	cache.setCapacity(config.cachesize);

	// cache.on[ins.start] = false;
	// cache.on[ins.end] = false;
}

void Counter::genkey(BitVec &key)
{
	vector<bool> tmpkey;

	int nodes = ins.nNodes();
	tmpkey.clear();
	tmpkey.resize(nodes, false);

	for(int i=0; i<nodeTrail.size(); i++)
		tmpkey[nodeTrail[i]] = true;

	key.setVec(tmpkey);
}

//=================================================================================================
// Inline Methods
// NOTE: These are copies of the inline methods in Glucose.

inline unsigned int Counter::computeLBDMC(const vec<Lit> & lits,int end) {
	int nblevels = 0;
	MYFLAG++;

	if(incremental) { // ----------------- INCREMENTAL MODE
		if(end==-1) end = lits.size();
		unsigned int nbDone = 0;
		for(int i=0;i<lits.size();i++) {
			if(nbDone>=end) break;
			if(isSelector(var(lits[i]))) continue;
			nbDone++;
			int l = level(var(lits[i]));
			if (permDiff[l] != MYFLAG) {
				permDiff[l] = MYFLAG;
				nblevels++;
			}
		}
	} else { // -------- DEFAULT MODE. NOT A LOT OF DIFFERENCES... BUT EASIER TO READ
		for(int i=0;i<lits.size();i++) {
			int l = level(var(lits[i]));
			if (permDiff[l] != MYFLAG) {
				permDiff[l] = MYFLAG;
				nblevels++;
			}
		}
	}

	return nblevels;
}

inline unsigned int Counter::computeLBDMC(const Clause &c) {
	int nblevels = 0;
	MYFLAG++;

	if(incremental) { // ----------------- INCREMENTAL MODE
		int nbDone = 0;
		for(int i=0;i<c.size();i++) {
			if(nbDone>=c.sizeWithoutSelectors()) break;
			if(isSelector(var(c[i]))) continue;
			nbDone++;
			int l = level(var(c[i]));
			if (permDiff[l] != MYFLAG) {
				permDiff[l] = MYFLAG;
				nblevels++;
			}
		}
	} else { // -------- DEFAULT MODE. NOT A LOT OF DIFFERENCES... BUT EASIER TO READ
		for(int i=0;i<c.size();i++) {
			int l = level(var(c[i]));
			if (permDiff[l] != MYFLAG) {
				permDiff[l] = MYFLAG;
				nblevels++;
			}
		}
	}
	return nblevels;
}
